import os
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import Any

from loguru import logger
from pydantic import AnyHttpUrl, BaseModel, ConfigDict, Field, SecretStr, field_validator, model_validator


@dataclass
class KonaGlobalState:
    root_path: Path


# TODO(es3n1n): this is very sketchy
kona_global_state = KonaGlobalState(root_path=Path.cwd())


class KonaEndpointType(StrEnum):
    HTTP = 'http'
    HTTPS = 'https'
    SOCAT = 'socat'
    NC = 'nc'
    NCAT_SSL = 'ncat-ssl'


class AttachmentAdditionalFile(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    path: str
    str_content: str | None = Field(default=None, alias='str')
    base64_content: str | None = Field(default=None, alias='base64')

    @model_validator(mode='after')
    def exactly_one_content(self) -> 'AttachmentAdditionalFile':
        has_str = self.str_content is not None
        has_b64 = self.base64_content is not None
        if has_str == has_b64:
            msg = 'exactly one of str or base64 must be provided'
            raise ValueError(msg)
        return self


class AttachmentConfig(BaseModel):
    files: list[str] = []
    exclude: list[str] = []
    additional: list[AttachmentAdditionalFile] = []
    pre_compressed: list[str] = []


class FlagValue(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    str_content: str | None = Field(default=None, alias='str')
    file: str | None = None

    @model_validator(mode='after')
    def exactly_one_source(self) -> 'FlagValue':
        has_str = self.str_content is not None
        has_file = self.file is not None
        if has_str == has_file:
            msg = 'exactly one of str or file must be provided'
            raise ValueError(msg)
        return self

    def resolve(self, challenge_dir: Path) -> str:
        if self.str_content is not None:
            return self.str_content
        if self.file is not None:
            return (challenge_dir / self.file).read_text().strip()
        raise RuntimeError


class KonaChallengeItem(BaseModel):
    class CTFD(BaseModel):
        class ChallengeState(StrEnum):
            VISIBLE = 'visible'
            HIDDEN = 'hidden'

        class Hint(BaseModel):
            hint: str
            cost: int = 0
            title: str | None = None

        state: ChallengeState = ChallengeState.VISIBLE
        type: str = 'dynamic'
        topics: list[str] = []
        tags: list[str] = []
        hints: list[Hint] = []
        connection_info: str = ''

    class Scoring(BaseModel):
        class CTFD(BaseModel):
            decay_function: str = 'logarithmic'
            decay: int = 60
            max_attempts: int = 0

        class RCTF(BaseModel):
            eligible_for_tiebreaks: bool = True

        ctfd: CTFD = CTFD()
        rctf: RCTF = RCTF()
        initial_value: int = 500
        minimum_value: int = 100

    class Flags(BaseModel):
        class CTFDFlag(BaseModel):
            type: str = 'static'
            flag: str | FlagValue

        rctf: str | FlagValue = ''
        ctfd: list[CTFDFlag] = []

    class InstancerConfig(BaseModel):
        class ExposeKind(StrEnum):
            TCP = 'tcp'
            TCP_SSL = 'tcp-ssl'
            HTTP = 'http'
            HTTPS = 'https'

        class Expose(BaseModel):
            kind: 'KonaChallengeItem.InstancerConfig.ExposeKind'
            host_prefix: str
            container_name: str
            container_port: int
            should_display: bool = True
            title: str | None = None

        challenge_integration_id: str
        config: Any = {}
        expose: list[Expose] = []
        timeout_milliseconds: int | None = None

    class Endpoint(BaseModel):
        name: str | None = None
        type: KonaEndpointType
        endpoint: str
        port: int | None = None

        @property
        def name_prefix(self) -> str:
            return f'{self.name}: ' if self.name else ''

        @property
        def http_port_if_needed(self) -> str:
            return f':{self.port}' if self.port else ''

        @property
        def http_endpoint(self) -> str:
            return f'{self.type.value}://{self.endpoint}{self.http_port_if_needed}'

    category: str
    name: str
    author: str
    override_id: str | None = None

    description: str = ''

    attachments: list[str] | AttachmentConfig = []

    scoring: Scoring = Scoring()
    flags: Flags = Flags()
    endpoints: list[Endpoint] = []

    ctfd: CTFD = CTFD()
    instancer_config: InstancerConfig | None = None

    @property
    def challenge_id(self) -> str:
        if self.override_id is not None:
            return self.override_id
        return f'{self.category}_{self.name}'

    @field_validator('description')
    @classmethod
    def strip_description(cls, v: str) -> str:
        return v.strip()

    @model_validator(mode='after')
    def flag_is_set(self) -> 'KonaChallengeItem':
        if self.flags.rctf or self.flags.ctfd:
            return self
        logger.warning(f'No flags set for challenge {self.challenge_id}')
        return self

    @model_validator(mode='after')
    def warn_attachments(self) -> 'KonaChallengeItem':
        has_attachments = (
            bool(self.attachments)
            if isinstance(self.attachments, list)
            else bool(self.attachments.files or self.attachments.pre_compressed)
        )
        if not has_attachments:
            logger.warning(f'No attachments set for challenge {self.challenge_id}')
        return self

    def resolve_flags(self, challenge_dir: Path) -> None:
        if isinstance(self.flags.rctf, FlagValue):
            self.flags.rctf = self.flags.rctf.resolve(challenge_dir)
        for flag in self.flags.ctfd:
            if isinstance(flag.flag, FlagValue):
                flag.flag = flag.flag.resolve(challenge_dir)


class KonaRolloutRestartConfig(BaseModel):
    annotation_path: str | None = None
    image: bool = True


class KonaChallengeConfig(BaseModel):
    class DiscoveryConfig(BaseModel):
        skip: bool = False

    class ChallengeDeploymentConfig(BaseModel):
        class DockerImage(BaseModel):
            path: str
            name: str
            tag: str = 'latest'
            registry_name: str | None = None
            build_args: dict[str, str] = {}
            platform: str | None = None
            no_cache: bool = False

        class KonaKubernetesManifest(BaseModel):
            paths: list[str]
            cluster_name: str | None = None
            rollout_restart: KonaRolloutRestartConfig = KonaRolloutRestartConfig()

        class KonaKubernetesInlineManifest(BaseModel):
            documents: list[dict[str, Any]]
            cluster_name: str | None = None
            rollout_restart: KonaRolloutRestartConfig = KonaRolloutRestartConfig()

        images: list[DockerImage] = []
        # TODO(es3n1n): rename kubernetes_inline_manifests to just kubernetes_manifests
        kubernetes_manifests: list[KonaKubernetesManifest] = []
        kubernetes_inline_manifests: list[KonaKubernetesInlineManifest] = []

    discovery: DiscoveryConfig = DiscoveryConfig()
    challenges: list[KonaChallengeItem] = []
    deployment: ChallengeDeploymentConfig = ChallengeDeploymentConfig()


class KonaSecret(BaseModel):
    file_path: str | None = None
    value: SecretStr | None = None
    env: str | None = None

    @model_validator(mode='after')
    def exactly_one_of(self) -> 'KonaSecret':
        provided = [f for f in ('file_path', 'value', 'env') if getattr(self, f) is not None]
        if len(provided) != 1:
            msg = 'exactly one of file_path, value, env must be provided'
            raise ValueError(msg)
        return self

    @property
    def loaded(self) -> str:
        if self.value is not None:
            return self.value.get_secret_value()

        if self.file_path is not None:
            file_path = Path(self.file_path)
            if self.file_path.startswith('.'):
                # TODO(es3n1n): move to an util
                file_path = (kona_global_state.root_path / self.file_path).resolve().absolute()

            if not file_path.exists():
                msg = f'{file_path} does not exist'
                raise FileNotFoundError(msg)

            self.value = SecretStr(file_path.read_text())
            return self.value.get_secret_value()

        if self.env is not None:
            value = os.getenv(self.env)
            if value is None:
                msg = f'Environment variable {self.env} is not set'
                raise ValueError(msg)
            self.value = SecretStr(value)
            return self.value.get_secret_value()

        raise RuntimeError


class KonaSecretOrValue(BaseModel):
    secret: str | None = None
    value: SecretStr | None = None

    @model_validator(mode='after')
    def exactly_one_of(self) -> 'KonaSecretOrValue':
        provided = [f for f in ('secret', 'value') if getattr(self, f) is not None]
        if len(provided) != 1:
            msg = 'exactly one of secret, value must be provided'
            raise ValueError(msg)
        return self

    def load(self, global_config: 'KonaGlobalConfig') -> str:
        if self.secret is not None:
            return global_config.secrets[self.secret].loaded
        if self.value is not None:
            return self.value.get_secret_value()
        raise RuntimeError


class KonaRCTFCredentials(BaseModel):
    base_url: AnyHttpUrl
    team_token: KonaSecretOrValue
    extra_headers: dict[str, str] = {}


class KonaCTFDCredentials(BaseModel):
    base_url: AnyHttpUrl
    admin_token: KonaSecretOrValue
    extra_headers: dict[str, str] = {}


class KonaDiscoveryConfig(BaseModel):
    challenge_folder_depth: int = 3
    attachment_analysis_depth: int = 50
    klodd_domain: str | None = None
    klodd_endpoint_name: str | None = None


class KonaTemplatesConfig(BaseModel):
    # TODO(es3n1n): inlining text here is ugly, consider loading from files
    challenge_description: str = (
        '{{ challenge.description }}\n\n{{ endpoints_rendered.strip() }}\n\n**Author**: {{ challenge.author }}'
    )
    endpoints_text: str = """{% for endpoint in challenge.endpoints -%}
{% if endpoint.type == models.KonaEndpointType.SOCAT %}
{{ endpoint.name_prefix }}`socat -,raw,echo=0 tcp:{{ endpoint.endpoint }}:{{ endpoint.port or 1337 }}`
{% elif endpoint.type == models.KonaEndpointType.NC %}
{{ endpoint.name_prefix }}`nc {{ endpoint.endpoint }} {{ endpoint.port or 1337 }}`
{% elif endpoint.type == models.KonaEndpointType.NCAT_SSL %}
{{ endpoint.name_prefix }}`ncat --ssl {{ endpoint.endpoint }} {{ endpoint.port or 1337 }}`
{% elif endpoint.type in (models.KonaEndpointType.HTTP, models.KonaEndpointType.HTTPS) %}
{% if endpoint.name -%}
[{{ endpoint.name }}]({{ endpoint.http_endpoint }})
{% else -%}
[{{ endpoint.http_endpoint }}]({{ endpoint.http_endpoint }})
{% endif -%}
{% else -%}
unknown endpoint type {{ endpoint.type }}
{% endif -%}
{% endfor -%}"""
    ctfd_attribution: str = '**Author**: {{ challenge.author }}'

    @field_validator('challenge_description', 'endpoints_text', 'ctfd_attribution')
    @classmethod
    def strip_values(cls, v: str) -> str:
        return v.strip()


class KonaGcloudClusterConfig(BaseModel):
    cluster_name: str
    project: str
    zone: str


class KonaKindClusterConfig(BaseModel):
    cluster_name: str = 'kind'


class KonaKubernetesClusterConfig(BaseModel):
    incluster: bool = False
    # KUBECONFIG or ~/.kube/config
    use_default: bool = True
    kubeconfig: KonaSecretOrValue | None = None
    gcloud: KonaGcloudClusterConfig | None = None
    kind: KonaKindClusterConfig | None = None
    alias_to: str | list[str] | None = None


class AttachmentFormat(StrEnum):
    TAR_GZ = 'tar_gz'
    ZIP = 'zip'


class KonaGlobalConfig(BaseModel):
    discovery: KonaDiscoveryConfig = KonaDiscoveryConfig()
    secrets: dict[str, KonaSecret] = {}
    rctf: KonaRCTFCredentials | None = None
    ctfd: KonaCTFDCredentials | None = None
    templates: KonaTemplatesConfig = KonaTemplatesConfig()
    registries: dict[str, str] = {}
    clusters: dict[str, KonaKubernetesClusterConfig] = {}
    domains: dict[str, str] = {}
    attachment_format: AttachmentFormat = AttachmentFormat.TAR_GZ



import base64
from fnmatch import fnmatch
from pathlib import Path

from kona.schema.models import AttachmentConfig, AttachmentFormat
from kona.util.tar import make_tar_gz_from
from kona.util.zip import make_zip


def _normalize_config(attachments: list[str] | AttachmentConfig) -> AttachmentConfig:
    if isinstance(attachments, list):
        return AttachmentConfig(files=attachments)
    return attachments


def _collect_paths(challenge_dir: Path, patterns: list[str]) -> list[Path]:
    result: list[Path] = []
    for pattern in patterns:
        path = challenge_dir / pattern
        if path.is_dir():
            result.extend(p for p in path.rglob('*') if p.is_file())
        elif any(c in pattern for c in ('*', '?', '[')):
            result.extend(p for p in challenge_dir.glob(pattern) if p.is_file())
        elif path.is_file():
            result.append(path)
    return result


def _is_excluded(rel_path: str, exclude_patterns: list[str]) -> bool:
    return any(fnmatch(rel_path, pat) for pat in exclude_patterns)


def resolve_source_paths(challenge_dir: Path, attachments: list[str] | AttachmentConfig) -> list[Path]:
    cfg = _normalize_config(attachments)
    paths = _collect_paths(challenge_dir, cfg.files)

    if cfg.exclude:
        paths = [p for p in paths if not _is_excluded(str(p.relative_to(challenge_dir)), cfg.exclude)]

    return paths


def resolve_attachments(
    challenge_dir: Path,
    tmp_dir: Path,
    attachments: list[str] | AttachmentConfig,
    fmt: AttachmentFormat,
    challenge_id: str,
) -> list[Path]:
    cfg = _normalize_config(attachments)
    result: list[Path] = []

    collected = _collect_paths(challenge_dir, cfg.files)

    if cfg.exclude:
        collected = [p for p in collected if not _is_excluded(str(p.relative_to(challenge_dir)), cfg.exclude)]

    entries: list[tuple[Path, str]] = [(p, str(p.relative_to(challenge_dir))) for p in collected]
    for additional in cfg.additional:
        dest = tmp_dir / additional.path
        dest.parent.mkdir(parents=True, exist_ok=True)
        if additional.str_content is not None:
            dest.write_text(additional.str_content)
        elif additional.base64_content is not None:
            dest.write_bytes(base64.b64decode(additional.base64_content))
        entries.append((dest, additional.path))

    if entries:
        archive_name = f'{challenge_id}.zip' if fmt == AttachmentFormat.ZIP else f'{challenge_id}.tar.gz'
        archive_path = tmp_dir / archive_name
        if fmt == AttachmentFormat.ZIP:
            make_zip(archive_path, entries)
        else:
            make_tar_gz_from(archive_path, entries)
        result.append(archive_path)

    for pre in cfg.pre_compressed:
        pre_path = challenge_dir / pre
        if pre_path.is_file():
            result.append(pre_path)

    return result



from typing import Any

from jinja2.sandbox import SandboxedEnvironment

from kona.schema import models


def render_template(template: str, **kwargs: Any) -> str:  # noqa: ANN401
    env = SandboxedEnvironment()
    tmpl = env.from_string(template)
    return tmpl.render(models=models, **kwargs)


def render_template_values(obj: Any, **kwargs: Any) -> Any:  # noqa: ANN401
    if isinstance(obj, str):
        return render_template(obj, **kwargs)
    if isinstance(obj, dict):
        return {k: render_template_values(v, **kwargs) for k, v in obj.items()}
    if isinstance(obj, list):
        return [render_template_values(v, **kwargs) for v in obj]
    return obj



from pathlib import Path

import click
from loguru import logger

from kona.analysis import include_passes
from kona.analysis.passes import AnalysisContext, passes
from kona.schema.models import AttachmentFormat
from kona.util.tar import make_tar_gz_from
from kona.util.zip import make_zip

from .core.sync import sync
from .schema.models import KonaGlobalConfig, kona_global_state
from .schema.parsers import load_schema


try:
    from uvloop import run  # type: ignore[import-not-found]
except ImportError:
    from asyncio import run  # type: ignore[no-redef]


async def job(deploy_directory: str, only: tuple[str, ...] = ()) -> None:
    kona_global_state.root_path = Path(deploy_directory).resolve().absolute()
    logger.info(f'Starting in {kona_global_state.root_path}')
    logger.info(f'Included {include_passes()} passes')

    only_challenges: tuple[str, ...] | None = only if only else None
    kona_config = load_schema(kona_global_state.root_path, model=KonaGlobalConfig)
    sync_result = await sync(kona_global_state.root_path, kona_config, only_challenges=only_challenges)

    context = AnalysisContext(
        global_config=kona_config,
        sync_result=sync_result,
    )
    for analysis_pass in passes:
        logger.info(f'Running pass {analysis_pass.__name__}')
        await analysis_pass(context)

    logger.info('We are done here')


@click.group()
def main() -> None:
    pass


@main.command()
@click.option(
    '-d',
    '--deploy-directory',
    'deploy_directory',
    type=click.Path(exists=True, file_okay=False),
    required=True,
)
@click.option(
    '--only',
    'only',
    multiple=True,
    help='Only sync specific challenge directories (paths relative to deploy-directory).',
)
@logger.catch(reraise=True)
def sync_cmd(deploy_directory: str, only: tuple[str, ...]) -> None:
    run(job(deploy_directory, only=only))


@main.command('compress')
@click.argument('path', type=click.Path(exists=True))
@click.option(
    '-f',
    '--format',
    'fmt',
    type=click.Choice(['tar_gz', 'zip'], case_sensitive=False),
    default='tar_gz',
    show_default=True,
)
@click.option(
    '-o',
    '--output',
    'output_path',
    type=click.Path(),
    default=None,
)
@logger.catch(reraise=True)
def compress_cmd(path: str, fmt: str, output_path: str | None) -> None:
    source = Path(path).resolve()
    attachment_fmt = AttachmentFormat(fmt)

    if output_path is None:
        ext = '.zip' if attachment_fmt == AttachmentFormat.ZIP else '.tar.gz'
        output = Path.cwd() / f'{source.name}{ext}'
    else:
        output = Path(output_path).resolve()

    if source.is_dir():
        all_files = [p for p in source.rglob('*') if p.is_file()]
        entries = [(p, str(p.relative_to(source))) for p in all_files]
    else:
        entries = [(source, source.name)]

    if not entries:
        logger.warning('No files to compress')
        return

    if attachment_fmt == AttachmentFormat.ZIP:
        make_zip(output, entries)
    else:
        make_tar_gz_from(output, entries)

    logger.info(f'Compressed {len(entries)} file(s) to {output}')


if __name__ == '__main__':
    main()



import os
import time

import psycopg2
from httpx import Client
from loguru import logger


RCTF_DATABASE_PASSWORD = os.getenv('RCTF_DATABASE_PASSWORD', 'kona')
RCTF_REDIS_PASSWORD = os.getenv('RCTF_REDIS_PASSWORD', 'kona')
RCTF_DATABASE_HOST = os.getenv('RCTF_DATABASE_HOST', 'postgres')
RCTF_REDIS_HOST = os.getenv('RCTF_REDIS_HOST', 'redis')
RCTF_HOST = os.getenv('RCTF_HOST', 'rctf')
HTTP_OK = 200


def main() -> None:
    logger.info('Waiting for rCTF to get up')

    with Client(base_url=f'http://{RCTF_HOST}', timeout=1) as client:
        while True:
            try:
                response = client.get('/api/v1/leaderboard/now?limit=100&offset=0')
                if response.status_code == HTTP_OK:
                    break
            except Exception as e:  # noqa: BLE001
                logger.warning(f'rCTF is not up yet: {e}')
            time.sleep(1)

        logger.info('rCTF is up')

        r = client.post(
            '/api/v1/auth/register',
            json={
                'name': 'admin',
                'email': 'admin@es3n1n.eu',
            },
        )
        if r.status_code != HTTP_OK:
            logger.error('Failed to register admin user')
            return

        logger.info('Created the user')

        conn = psycopg2.connect(
            f"dbname='kona' user='kona' host='{RCTF_DATABASE_HOST}' password='{RCTF_DATABASE_PASSWORD}'"
        )
        with conn.cursor() as cursor:
            # This way, our team token will be persistent across all restarts
            cursor.execute(
                'UPDATE users SET '
                "id = '70aba09d-37fe-44b1-b9d8-531efd3d5743', "
                'perms = 7, '
                "created_at = '2025-09-19 15:05:05.590029+00'"
            )
            conn.commit()


if __name__ == '__main__':
    main()



import os
from time import sleep

import pymysql
from httpx import Client
from loguru import logger


CTFD_HOST = os.getenv('CTFD_HOST', 'ctfd')
MYSQL_HOST = os.getenv('MYSQL_HOST', 'mysql')
HTTP_OK = 200
HTTP_REDIRECT = 302


def main() -> None:
    nonce: str
    with Client(base_url=f'http://{CTFD_HOST}:8000', timeout=1) as client:
        while True:
            try:
                r = client.get('/', follow_redirects=True)
                if r.status_code in (HTTP_OK, HTTP_REDIRECT):
                    nonce = r.text.split("'csrfNonce': \"")[1].split('"')[0]
                    logger.info(f'Got nonce: {nonce}')
                    break
            except Exception as e:  # noqa: BLE001
                logger.warning(f'CTFd is not up yet: {e}')
            sleep(2)

        r = client.post(
            '/setup',
            data={
                'ctf_name': 'kona',
                'ctf_description': 'kona',
                'user_mode': 'teams',
                'challenge_visibility': 'private',
                'account_visibility': 'public',
                'score_visibility': 'public',
                'registration_visibility': 'public',
                'verify_emails': 'false',
                'team_size': '',
                'name': 'admin',
                'email': 'admin@es3n1n.eu',
                'password': 'admin',
                'ctf_logo': '',
                'ctf_banner': '',
                'ctf_small_icon': '',
                'ctf_theme': 'core',
                'theme_color': '',
                'start': '79860',
                'end': '25466523060',
                'social_shares': 'true',
                '_submit': 'Finish',
                'nonce': nonce,
            },
            follow_redirects=True,
        )
        r.raise_for_status()
        logger.info(f'CTFd setup response: {r.status_code}')

        if 'admin@es3n1n.eu' not in r.text:
            msg = f'CTFd setup failed: {r.text}'
            raise ValueError(msg)

    with pymysql.connect(host=MYSQL_HOST, user='ctfd', password='ctfd', database='ctfd') as conn:  # noqa: S106
        with conn.cursor() as cursor:
            # This way, access token will be persistent across all restarts
            cursor.execute(
                'INSERT INTO `tokens`(type, user_id, created, expiration, value, description) VALUES '
                "('user', 1, '2025-09-19 21:58:10.008657', '9999-01-01 00:00:00.000000', "
                "'ctfd_5bbb659f3ffb85533260c40af3fd46cd0ca4a742fe80fc7dac366e38c4675424', 'kona')",
            )
        conn.commit()

    logger.info('CTFd setup completed')


if __name__ == '__main__':
    main()



import tomllib
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import yaml
from pydantic import TypeAdapter


if TYPE_CHECKING:
    from collections.abc import Callable

    from pydantic import BaseModel

T = TypeVar('T', bound='BaseModel')


def load_yaml[T](file_path: Path, model: type[T]) -> T:
    return TypeAdapter(model).validate_python(
        yaml.safe_load(file_path.read_bytes()),
    )


def load_toml[T](file_path: Path, model: type[T]) -> T:
    return TypeAdapter(model).validate_python(
        tomllib.loads(file_path.read_bytes().decode('utf-8')),
    )


def load_schema[T](
    directory: Path,
    *,
    file_name: str = 'kona',
    model: type[T],
) -> T:
    for ext in ('yaml', 'yml', 'toml'):
        loader: Callable[[Path, type[T]], T] = load_yaml if ext in ('yaml', 'yml') else load_toml
        file_path = directory / f'{file_name}.{ext}'

        if not file_path.exists():
            continue

        return loader(file_path, model)

    msg = f'No supported schema file found for {file_name} in {directory}'
    raise FileNotFoundError(msg)


def try_load_schema[T](
    directory: Path,
    *,
    file_name: str = 'kona',
    model: type[T],
) -> T | None:
    try:
        return load_schema(directory, file_name=file_name, model=model)
    except FileNotFoundError:
        return None



import zipfile
from pathlib import Path


_FIXED_DATE_TIME = (1980, 1, 1, 0, 0, 0)


def make_zip(output_path: Path, source_files: list[tuple[Path, str]]) -> None:
    with zipfile.ZipFile(str(output_path.absolute()), 'w', compression=zipfile.ZIP_DEFLATED) as zf:
        for fs_path, arcname in sorted(source_files, key=lambda x: x[1]):
            info = zipfile.ZipInfo(filename=arcname, date_time=_FIXED_DATE_TIME)
            info.external_attr = 0o777 << 16
            zf.writestr(info, fs_path.read_bytes())



import importlib
from pathlib import Path


def include_passes() -> int:
    result = 0
    for p in Path(__file__).parent.glob('*.py'):
        if p.stem in {'__init__', 'passes'}:
            continue

        importlib.import_module(f'kona.analysis.{p.stem}')
        result += 1

    return result



import io
import subprocess
from dataclasses import dataclass

from kubernetes.config import load_incluster_config, load_kube_config
from loguru import logger

from kona.schema.models import KonaGlobalConfig


@dataclass
class KubernetesState:
    last_loaded_cluster: str | None = None


kubernetes_state = KubernetesState()


def _run_checked(args: list[str]) -> str:
    result = subprocess.run(args, capture_output=True, text=True, check=False)  # noqa: S603
    if result.returncode != 0:
        output = (result.stdout + result.stderr).strip()
        msg = f'Command {args} failed (exit {result.returncode}): {output}'
        raise RuntimeError(msg)
    return result.stdout


def _load_gcloud_credentials(cluster_name: str, project: str, zone: str) -> None:
    _run_checked(
        [
            'gcloud',
            'container',
            'clusters',
            'get-credentials',
            cluster_name,
            '--project',
            project,
            '--zone',
            zone,
        ]
    )
    logger.info(f'Loaded gcloud credentials for cluster "{cluster_name}" in {project}/{zone}')


def _load_kind_credentials(cluster_name: str) -> None:
    _run_checked(['kind', 'export', 'kubeconfig', '--name', cluster_name])
    logger.info(f'Loaded kind credentials for cluster "{cluster_name}"')


def resolve_cluster_names(global_config: KonaGlobalConfig, cluster_name: str) -> list[str]:
    cluster = global_config.clusters.get(cluster_name)
    if cluster is None:
        msg = f'Unknown cluster "{cluster_name}"'
        raise ValueError(msg)

    if cluster.alias_to is None:
        return [cluster_name]

    targets = [cluster.alias_to] if isinstance(cluster.alias_to, str) else cluster.alias_to
    resolved: list[str] = []
    for target in targets:
        resolved.extend(resolve_cluster_names(global_config, target))
    return resolved


def _load_kubeconfig_single(global_config: KonaGlobalConfig, cluster_name: str) -> None:
    cluster = global_config.clusters.get(cluster_name)
    if cluster is None:
        msg = f'Unknown cluster "{cluster_name}"'
        raise ValueError(msg)

    if cluster.alias_to is not None:
        msg = f'Cluster "{cluster_name}" is an alias, resolve it first'
        raise ValueError(msg)

    if cluster.gcloud:
        _load_gcloud_credentials(cluster.gcloud.cluster_name, cluster.gcloud.project, cluster.gcloud.zone)
        load_kube_config()
        return

    if cluster.kind:
        _load_kind_credentials(cluster.kind.cluster_name)
        load_kube_config()
        return

    if not cluster.kubeconfig:
        if cluster.incluster:
            load_incluster_config()
            logger.info(f'Loaded incluster config for cluster "{cluster_name}"')
            return

        if cluster.use_default:
            load_kube_config()
            logger.info(f'Loaded default config for cluster "{cluster_name}"')
            return

        msg = f'Unable to load config for cluster "{cluster_name}"'
        raise ValueError(msg)

    kubeconfig = cluster.kubeconfig.load(global_config)
    load_kube_config(
        config_file=io.BytesIO(kubeconfig.encode()),
    )
    logger.info(f'Loaded kubeconfig for cluster "{cluster_name}"')


def load_kubeconfig(global_config: KonaGlobalConfig, cluster_name: str) -> None:
    if kubernetes_state.last_loaded_cluster is not None and kubernetes_state.last_loaded_cluster == cluster_name:
        return

    _load_kubeconfig_single(global_config, cluster_name)
    kubernetes_state.last_loaded_cluster = cluster_name



from httpx import Response


def raise_for_status(response: Response) -> None:
    if response.is_success:
        return
    msg = f'{response.status_code} {response.reason_phrase} for url {response.url}: {response.text}'
    raise RuntimeError(msg)



from loguru import logger

from kona.schema.models import KonaChallengeConfig, KonaChallengeItem, KonaEndpointType, KonaGlobalConfig

from .deployment import DeploymentResult


def _find_manifests(api_version: str, kind: str, manifests: list[dict]) -> list[dict]:
    return list(
        filter(lambda manifest: manifest.get('kind') == kind and manifest.get('apiVersion') == api_version, manifests)
    )


def discover_klodd_endpoint(
    config: KonaGlobalConfig, challenge: KonaChallengeConfig, deployment_result: DeploymentResult
) -> None:
    for klodd_challenge in _find_manifests(
        api_version='klodd.tjcsec.club/v1',
        kind='Challenge',
        manifests=deployment_result.deployed_kubernetes_manifests,
    ):
        metadata = klodd_challenge.get('metadata', {})
        challenge_name = metadata.get('name')
        if not challenge_name:
            logger.warning(f'Unable to find challenge name for {klodd_challenge=}')
            continue

        if not config.discovery.klodd_domain:
            logger.warning(f'Auto-discovered klodd challenge {challenge_name}, but klodd domain is not set!')
            continue

        endpoint = f'{config.discovery.klodd_domain}/challenge/{challenge_name}'
        logger.info(f'Discovered {challenge_name} starter endpoint {endpoint}')

        for item in challenge.challenges:
            item.endpoints.append(
                KonaChallengeItem.Endpoint(
                    name=config.discovery.klodd_endpoint_name,
                    type=KonaEndpointType.HTTPS,
                    endpoint=endpoint,
                )
            )


def discover_deployed_endpoints(
    config: KonaGlobalConfig, challenge: KonaChallengeConfig, deployment_result: DeploymentResult
) -> None:
    discover_klodd_endpoint(config, challenge, deployment_result)



from abc import ABC, abstractmethod
from pathlib import Path

from kona.schema.models import KonaChallengeItem


class ExternalProviderABC(ABC):
    @abstractmethod
    async def setup(self) -> None:
        pass

    @abstractmethod
    async def sync_challenge(
        self, challenge: KonaChallengeItem, attachment_paths: list[Path], rendered_description: str
    ) -> None:
        pass



import hashlib
from pathlib import Path

from httpx import AsyncClient
from loguru import logger

from kona.schema.models import KonaChallengeItem, KonaCTFDCredentials, KonaGlobalConfig
from kona.util.http import raise_for_status
from kona.util.jinja import render_template

from .abc import ExternalProviderABC


def flag_to_ctfd(challenge_id: int, flag: KonaChallengeItem.Flags.CTFDFlag) -> dict:
    return {
        'content': flag.flag,
        'type': flag.type,
        'challenge_id': challenge_id,
    }


def topic_to_ctfd(challenge_id: int, topic: str) -> dict:
    return {
        'value': topic,
        'type': 'challenge',
        'challenge_id': challenge_id,
    }


def tag_to_ctfd(challenge_id: int, tag: str) -> dict:
    return {
        'value': tag,
        'challenge_id': challenge_id,
    }


def hint_to_ctfd(challenge_id: int, hint: KonaChallengeItem.CTFD.Hint) -> dict:
    return {
        'content': hint.hint,
        'title': hint.title,
        'cost': hint.cost,
        'challenge_id': challenge_id,
    }


def filter_items(
    on_server: list[dict], local: list[dict], keys: list[str] | None = None
) -> tuple[list[dict], list[dict]]:
    to_add: list[dict] = []
    to_remove: list[dict] = []

    if not local:
        return to_add, on_server

    if not on_server:
        return local, to_remove

    if keys is None:
        keys = list(local[0].keys())

    to_add = [
        local_item
        for local_item in local
        if not any(all(local_item[k] == server_item[k] for k in keys) for server_item in on_server)
    ]
    to_remove = [
        server_item
        for server_item in on_server
        if not any(all(server_item[k] == local_item[k] for k in keys) for local_item in local)
    ]
    return to_add, to_remove


class CTFDProvider(ExternalProviderABC):
    def __init__(self, global_config: KonaGlobalConfig, credentials: KonaCTFDCredentials) -> None:
        self.global_config = global_config
        self.credentials = credentials
        self.admin_token: str = credentials.admin_token.load(global_config=global_config)
        self.challenges_on_remote: list[dict] = []

    @property
    def _client(self) -> AsyncClient:
        return AsyncClient(
            base_url=str(self.credentials.base_url),
            headers={
                **self.credentials.extra_headers,
                'Authorization': f'Token {self.admin_token}',
                'Content-Type': 'application/json',  # very important, lol...
            },
        )

    async def setup(self) -> None:
        async with self._client as client:
            r = await client.get('/api/v1/challenges?view=admin')
            raise_for_status(r)
            self.challenges_on_remote = r.json()['data']
        logger.info(f'Retrieved {len(self.challenges_on_remote)} CTFd challenges from remote')

    async def _load_full_challenge(self, challenge_id: int) -> dict:
        async with self._client as client:
            r = await client.get(f'/api/v1/challenges/{challenge_id}?view=admin')
            raise_for_status(r)
            return r.json()['data']

    async def _create_challenge(
        self, challenge: KonaChallengeItem, attachment_paths: list[Path], challenge_dict: dict
    ) -> None:
        async with self._client as client:
            r = await client.post(
                '/api/v1/challenges',
                json=challenge_dict,
            )
            raise_for_status(r)
            challenge_id = r.json()['data']['id']
            logger.info(f'Created challenge {challenge.name} in CTFd with ID {challenge_id}')

            # flags
            for flag in challenge.flags.ctfd:
                r = await client.post(
                    '/api/v1/flags',
                    json=flag_to_ctfd(challenge_id, flag),
                )
                raise_for_status(r)

            # topics
            for topic in challenge.ctfd.topics:
                r = await client.post(
                    '/api/v1/topics',
                    json=topic_to_ctfd(challenge_id, topic),
                )
                raise_for_status(r)

            # tags
            for tag in challenge.ctfd.tags:
                r = await client.post(
                    '/api/v1/tags',
                    json=tag_to_ctfd(challenge_id, tag),
                )
                raise_for_status(r)

            # hints
            for hint in challenge.ctfd.hints:
                r = await client.post(
                    '/api/v1/hints',
                    json=hint_to_ctfd(challenge_id, hint),
                )
                raise_for_status(r)

        # This **needs** to be without Content-Type header, let httpx set it properly with multipart boundary
        async with self._client as client:
            client.headers.pop('Content-Type', None)

            # files
            for attachment_path in attachment_paths:
                r = await client.post(
                    '/api/v1/files',
                    data={
                        'challenge_id': challenge_id,
                        'type': 'challenge',
                    },
                    files={'file': (attachment_path.name, attachment_path.read_bytes())},
                )
                raise_for_status(r)
                if len(r.json()['data']) < 1:
                    raise RuntimeError

    async def _update_challenge(
        self, challenge: KonaChallengeItem, attachment_paths: list[Path], challenge_dict: dict, current_challenge: dict
    ) -> None:
        # Initial patch if needed
        if any(challenge_dict[x] != current_challenge[x] for x in challenge_dict):
            async with self._client as client:
                r = await client.patch(
                    f'/api/v1/challenges/{current_challenge["id"]}',
                    json=challenge_dict,
                )
                raise_for_status(r)
                logger.info(f'Updated challenge {challenge.challenge_id} in CTFd with ID {current_challenge["id"]}')

        async def generic_updater_inner(
            name: str,
            _client: AsyncClient,
            add_url: str,
            remove_url: str,
            remote_items: list[dict],
            local_items: list[dict],
            keys: list[str] | None = None,
        ) -> None:
            add_items, remove_items = filter_items(
                remote_items,
                local_items,
                keys=keys,
            )

            for a_item in add_items:
                g_r = await _client.post(
                    add_url,
                    json=a_item,
                )
                raise_for_status(g_r)
                logger.info(
                    f'Added new {name} {a_item} to {challenge.challenge_id} in CTFd with ID {current_challenge["id"]}'
                )

            for a_item in remove_items:
                g_r = await _client.delete(remove_url.format(id=a_item['id']))
                raise_for_status(g_r)
                logger.info(
                    f'Removed {name} {a_item} from {challenge.challenge_id} in CTFd with ID {current_challenge["id"]}'
                )

        async def generic_updater(
            name: str,
            get_url: str,
            add_url: str,
            remove_url: str,
            local_items: list[dict],
            keys: list[str] | None = None,
        ) -> None:
            async with self._client as _client:
                g_r = await _client.get(get_url)
                raise_for_status(g_r)
                await generic_updater_inner(name, _client, add_url, remove_url, g_r.json()['data'], local_items, keys)

        # flags
        await generic_updater(
            name='flag',
            get_url=f'/api/v1/flags?challenge_id={current_challenge["id"]}',
            add_url='/api/v1/flags',
            remove_url='/api/v1/flags/{id}',
            local_items=[flag_to_ctfd(current_challenge['id'], flag) for flag in challenge.flags.ctfd],
        )

        # topics
        await generic_updater(
            name='topic',
            get_url=f'/api/v1/topics?challenge_id={current_challenge["id"]}',
            add_url='/api/v1/topics',
            remove_url='/api/v1/topics/{id}',
            local_items=[topic_to_ctfd(current_challenge['id'], topic) for topic in challenge.ctfd.topics],
            keys=['value'],
        )

        # tags
        await generic_updater(
            name='tag',
            get_url=f'/api/v1/tags?challenge_id={current_challenge["id"]}',
            add_url='/api/v1/tags',
            remove_url='/api/v1/tags/{id}',
            local_items=[tag_to_ctfd(current_challenge['id'], tag) for tag in challenge.ctfd.tags],
            keys=['value'],
        )

        # hints are a bit special
        async with self._client as _client:
            r = await _client.get(f'/api/v1/hints?challenge_id={current_challenge["id"]}')
            raise_for_status(r)
            server_hints: list[dict] = r.json()['data']
            for item in server_hints:
                r = await _client.get(f'/api/v1/hints/{item["id"]}?preview=true')
                raise_for_status(r)
                item['content'] = r.json()['data']['content']

            await generic_updater_inner(
                name='hint',
                _client=_client,
                add_url='/api/v1/hints',
                remove_url='/api/v1/hints/{id}',
                remote_items=server_hints,
                local_items=[hint_to_ctfd(current_challenge['id'], hint) for hint in challenge.ctfd.hints],
                keys=['content', 'title', 'cost'],
            )

        # files are special too
        async with self._client as _client:
            r = await _client.get(f'/api/v1/challenges/{current_challenge["id"]}/files')
            raise_for_status(r)

            server_files: list[dict] = r.json()['data']
            local_files: list[dict] = [
                {
                    'sha1sum': hashlib.sha1(ap.read_bytes(), usedforsecurity=False).hexdigest(),
                    'name': ap.name,
                    'data': ap.read_bytes(),
                }
                for ap in attachment_paths
            ]

            add_files, remove_files = filter_items(server_files, local_files, keys=['sha1sum'])
            for remove_file in remove_files:
                r = await _client.delete(f'/api/v1/files/{remove_file["id"]}')
                raise_for_status(r)
                logger.info(
                    f'Removed file {remove_file} from {challenge.challenge_id} in CTFd '
                    f'with ID {current_challenge["id"]}'
                )

            # This **needs** to be without Content-Type header, let httpx set it properly with multipart boundary
            _client.headers.pop('Content-Type', None)
            for add_file in add_files:
                r = await _client.post(
                    '/api/v1/files',
                    data={
                        'challenge_id': current_challenge['id'],
                        'type': 'challenge',
                    },
                    files={'file': (add_file['name'], add_file['data'])},
                )
                raise_for_status(r)
                logger.info(
                    f'Added file {add_file["name"]} to {challenge.challenge_id} in CTFd '
                    f'with ID {current_challenge["id"]}'
                )

    async def sync_challenge(
        self, challenge: KonaChallengeItem, attachment_paths: list[Path], rendered_description: str
    ) -> None:
        challenge_dict = {
            'name': challenge.name,
            'category': challenge.category,
            'description': rendered_description,
            'attribution': render_template(self.global_config.templates.ctfd_attribution, challenge=challenge),
            'type': challenge.ctfd.type,
            'state': challenge.ctfd.state.value,
            'max_attempts': challenge.scoring.ctfd.max_attempts,
            'connection_info': challenge.ctfd.connection_info,
        }

        if challenge.ctfd.type == 'static':
            challenge_dict['value'] = challenge.scoring.initial_value
        else:
            # TODO(es3n1n): this is wrong, need to think how to properly distinguish dynamic challenges
            challenge_dict['initial'] = challenge.scoring.initial_value
            challenge_dict['decay'] = challenge.scoring.ctfd.decay
            challenge_dict['minimum'] = challenge.scoring.minimum_value
            challenge_dict['function'] = challenge.scoring.ctfd.decay_function

        try:
            existing_challenge = next(
                chal
                for chal in self.challenges_on_remote
                if chal['name'] == challenge.name and chal['category'] == challenge.category
            )
            existing_challenge = await self._load_full_challenge(existing_challenge['id'])
        except StopIteration:
            existing_challenge = None

        if existing_challenge:
            logger.info(f'Challenge {challenge.challenge_id} already exists in CTFd with ID {existing_challenge["id"]}')
            await self._update_challenge(challenge, attachment_paths, challenge_dict, existing_challenge)
            return

        logger.info(f'Challenge {challenge.challenge_id} does not exist in CTFd, creating it')
        await self._create_challenge(challenge, attachment_paths, challenge_dict)



import gzip
import tarfile
from io import BytesIO
from pathlib import Path


def _norm_ti(ti: tarfile.TarInfo) -> tarfile.TarInfo:
    ti.uid = 0
    ti.gid = 0
    ti.uname = 'kona'
    ti.gname = 'kona'
    ti.mtime = 0
    ti.pax_headers = {}
    ti.mode = 0o777
    return ti


def _open_deterministic_gzip(output_path: Path) -> gzip.GzipFile:
    return gzip.GzipFile(filename='', mode='wb', mtime=0, fileobj=output_path.open('wb'))


def make_tar_gz(output_path: Path, source_files: list[Path]) -> None:
    with (
        _open_deterministic_gzip(output_path) as gz,
        tarfile.open(fileobj=gz, mode='w', format=tarfile.USTAR_FORMAT) as tar,
    ):
        for source_file in source_files:
            tar.add(source_file, arcname=source_file.name, filter=_norm_ti)


def make_tar_gz_from(output_path: Path, source_files: list[tuple[Path, str]]) -> None:
    with (
        _open_deterministic_gzip(output_path) as gz,
        tarfile.open(fileobj=gz, mode='w', format=tarfile.USTAR_FORMAT) as tar,
    ):
        for fs_path, arcname in sorted(source_files, key=lambda x: x[1]):
            data = fs_path.read_bytes()
            ti = tarfile.TarInfo(name=arcname)
            ti.size = len(data)
            ti = _norm_ti(ti)
            tar.addfile(ti, BytesIO(data))



from collections import defaultdict
from enum import StrEnum
from pathlib import Path

from loguru import logger

from kona.schema.models import KonaGlobalConfig, kona_global_state

from .passes import AnalysisContext, analysis_pass


class CountedDifferencesType(StrEnum):
    LINES = 'lines'
    BYTES = 'bytes'


def looks_like_text(file_data: bytes) -> bool:
    return any(not (ord(' ') <= x <= ord('~')) for x in file_data[:100])


def generic_diff(left: list[str] | bytes | str, right: list[str] | bytes | str) -> int:
    result = 0

    for i in range(max(len(left), len(right))):
        left_line = left[i] if i < len(left) else None
        right_line = right[i] if i < len(right) else None

        if left_line == right_line:
            continue

        result += 1

    return result


def diff(left: bytes, right: bytes) -> tuple[CountedDifferencesType, int]:
    if looks_like_text(left) and looks_like_text(right):
        return CountedDifferencesType.LINES, generic_diff(
            left.decode('utf-8', errors='ignore').splitlines(), right.decode('utf-8', errors='ignore').splitlines()
        )
    return CountedDifferencesType.BYTES, generic_diff(left, right)


def build_files_dict(
    global_config: KonaGlobalConfig, result: defaultdict[str, list[Path]], *items: Path, depth: int = 0
) -> defaultdict[str, list[Path]]:
    if depth > global_config.discovery.attachment_analysis_depth:
        return result

    for item in items:
        if item.is_dir():
            build_files_dict(global_config, result, *list(item.iterdir()), depth=depth + 1)
            continue

        result[item.name].append(item.resolve().absolute())

    return result


@analysis_pass
async def analyze_attachment(context: AnalysisContext) -> None:
    for group in context.sync_result.groups:
        attachments_dict = build_files_dict(
            context.global_config,
            defaultdict(list),
            *(attachment for chal in group.challenges for attachment in chal.attachments),
        )

        for image in group.deployment_result.built_docker_images:
            # Let's see what files are built for the docker image and if they are in attachments
            image_dict = build_files_dict(context.global_config, defaultdict(list), image.path)

            for file, image_paths in image_dict.items():
                attachment_files = attachments_dict.get(file)
                if not attachment_files:
                    continue

                for attachment_file, image_file in zip(attachment_files, image_paths, strict=False):
                    diff_type, diff_result = diff(image_file.read_bytes(), attachment_file.read_bytes())
                    if not diff_result:
                        # same content
                        continue

                    img_rel = image_file.relative_to(kona_global_state.root_path)
                    attachment_rel = attachment_file.relative_to(kona_global_state.root_path)
                    logger.warning(
                        f'Difference between {img_rel} and {attachment_rel} is {diff_result} {diff_type.name}'
                    )



import asyncio
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import UTC, datetime
from functools import cache
from http import HTTPStatus
from pathlib import Path
from typing import Any

import docker
import yaml
from kubernetes.client import ApiClient, ApiException
from kubernetes.dynamic import DynamicClient
from kubernetes.dynamic.exceptions import ResourceNotFoundError
from loguru import logger

from kona.schema.models import KonaChallengeConfig, KonaGlobalConfig
from kona.util.jinja import render_template, render_template_values

from .kubernetes import load_kubeconfig, resolve_cluster_names


DEFAULT_NAMESPACE = 'default'


@dataclass(frozen=True)
class BuiltDockerImage:
    path: Path
    full_ref: str
    digest: str | None = None


@dataclass
class DeploymentResult:
    deployed_kubernetes_manifests: list[dict] = field(default_factory=list)
    built_docker_images: list[BuiltDockerImage] = field(default_factory=list)


@cache
def docker_env() -> docker.DockerClient:
    return docker.from_env()


def docker_build_image(
    env: docker.DockerClient,
    context_dir: Path,
    full_ref: str,
    build_args: dict,
    platform: str | None,
    *,
    no_cache: bool,
) -> None:
    _, logs = env.images.build(
        path=str(context_dir),
        tag=full_ref,
        nocache=no_cache,
        pull=True,
        forcerm=True,
        buildargs=build_args,
        platform=platform,
    )
    for line in logs:
        logger.debug(str(line).strip())


def docker_push_image(env: docker.DockerClient, repository: str, tag: str) -> str | None:
    # First try without auth_config so credential helpers (e.g. gcloud) are used.
    # Fall back to auth_config={} for anonymous pushes — needed since Docker 28.3.3
    # which rejects pushes without an X-Registry-Auth header.
    # https://github.com/docker/docker-py/issues/3348#issuecomment-3224418755
    push_kwargs: dict[str, Any] = {}

    for attempt in range(2):
        error: str | None = None
        digest: str | None = None
        for line in env.images.push(
            repository=repository,
            tag=tag,
            stream=True,
            decode=True,
            **push_kwargs,
        ):
            if 'error' in line:
                error = line['error']
                break
            if 'aux' in line:
                digest = line['aux'].get('Digest')
            logger.debug(str(line).strip())

        if error is None:
            return digest

        if attempt == 0 and 'IncompleteRead' in error:
            logger.debug('Retrying push with empty auth_config (Docker 28.3.3+ workaround)')
            push_kwargs['auth_config'] = {}
            continue

        raise RuntimeError(error)
    return None


async def docker_build_images(
    result: DeploymentResult,
    config: KonaGlobalConfig,
    path: Path,
    deployment_config: KonaChallengeConfig.ChallengeDeploymentConfig,
) -> None:
    if not deployment_config.images:
        # No need to call docker_env if we don't need it
        return

    env = docker_env()
    for image in deployment_config.images:
        repository = image.name
        if image.registry_name:
            registry = config.registries.get(image.registry_name)
            if not registry:
                msg = f'Unknown registry name "{image.registry_name}" for {image.name}:{image.tag}'
                raise ValueError(msg)

            repository = f'{registry.rstrip("/")}/{image.name}'

        full_ref = f'{repository}:{image.tag}'
        full_path = (path / image.path).resolve().absolute()

        logger.info(f'Building {full_ref} for {image.name}:{image.tag}')

        await asyncio.to_thread(
            docker_build_image,
            env=env,
            context_dir=full_path,
            full_ref=full_ref,
            build_args=image.build_args,
            platform=image.platform,
            no_cache=image.no_cache,
        )

        digest: str | None = None
        if not image.registry_name:
            logger.warning(f'Skipping push for {full_ref} (no registry specified)')
        else:
            logger.info(f'Pushing {full_ref} for {image.name}:{image.tag}')
            digest = await asyncio.to_thread(docker_push_image, env=env, repository=repository, tag=image.tag)

        result.built_docker_images.append(
            BuiltDockerImage(
                path=full_path,
                full_ref=full_ref,
                digest=digest,
            )
        )


def _to_dict(obj: Any) -> dict[str, Any]:  # noqa: ANN401
    return obj.to_dict() if hasattr(obj, 'to_dict') else obj


def build_manifest_context(
    config: KonaGlobalConfig,
    result: DeploymentResult,
    deployment_config: KonaChallengeConfig.ChallengeDeploymentConfig,
    challenges: list,
    *,
    use_image_digest: bool = True,
) -> dict[str, Any]:
    images: dict[str, str] = {}
    for image, built in zip(deployment_config.images, result.built_docker_images, strict=False):
        ref = built.full_ref
        if use_image_digest and built.digest:
            ref = f'{ref}@{built.digest}'
        images[image.name] = ref
    return {
        'registries': config.registries,
        'images': images,
        'config': config,
        'challenges': challenges,
    }


def _k8s_get_resource(dyn: DynamicClient, api_version: str, kind: str) -> Any:  # noqa: ANN401
    try:
        return dyn.resources.get(api_version=api_version, kind=kind)
    except ResourceNotFoundError as err:
        msg = f'Unknown resource {api_version}/{kind}. Is the CRD installed?'
        raise ValueError(msg) from err


def _k8s_apply_manifest(
    dyn: DynamicClient,
    document: dict[str, Any],
) -> tuple[str, dict[str, Any]]:
    api_version = document.get('apiVersion')
    kind = document.get('kind')
    meta = document.get('metadata') or {}
    name = meta.get('name')
    if not api_version or not kind or not name:
        msg = 'manifest must include apiVersion, kind, metadata.name'
        raise ValueError(msg)

    resource = _k8s_get_resource(dyn, api_version, kind)
    kwargs: dict[str, Any] = {}
    if getattr(resource, 'namespaced', False):
        kwargs['namespace'] = meta.get('namespace') or DEFAULT_NAMESPACE

    try:
        resource.get(name=name, **kwargs)
    except ApiException as e:
        if e.status == HTTPStatus.NOT_FOUND:
            created = resource.create(body=document, **kwargs)
            return 'created', _to_dict(created)
        raise

    # strategic merge patch
    patched = resource.patch(
        name=name,
        body=document,
        content_type='application/merge-patch+json',
        **kwargs,
    )
    return 'configured', _to_dict(patched)


def k8s_expand_manifest(doc: dict[str, Any]) -> Iterable[dict[str, Any]]:
    if doc.get('kind') == 'List' and isinstance(doc.get('items'), list):
        return (i for i in doc['items'] if isinstance(i, dict))
    return (doc,)


def _resolve_cluster_targets(config: KonaGlobalConfig, cluster_name: str | None) -> list[str]:
    clusters_count = len(config.clusters)
    if not cluster_name:
        if clusters_count != 1:
            msg = 'Cluster name should always be defined when there are more than one cluster defined'
            raise ValueError(msg)
        return []
    return resolve_cluster_names(config, cluster_name)


async def _k8s_apply_items_to_cluster(
    result: DeploymentResult,
    config: KonaGlobalConfig,
    cluster_name: str | None,
    items: list[dict[str, Any]],
) -> None:
    if cluster_name:
        load_kubeconfig(config, cluster_name)

    dyn = DynamicClient(ApiClient())

    for item in items:
        try:
            action, obj = _k8s_apply_manifest(dyn, item)
        except Exception as e:
            ctx = f'{item.get("apiVersion")}/{item.get("kind")} {item.get("metadata", {}).get("name")}'
            logger.exception(f'Failed {ctx}: {e}')
            raise

        meta = (obj or item).get('metadata', {})
        api_version = item.get('apiVersion')
        kind = item.get('kind')
        name = meta.get('name', '<unnamed>')
        ns = meta.get('namespace')
        scope = f'{ns}/' if ns else ''
        logger.info(f'k8s {api_version}/{kind}:{scope}{name} {action}')

        enriched = dict(item)
        enriched['_action'] = action
        result.deployed_kubernetes_manifests.append(enriched)


async def _k8s_apply_items(
    result: DeploymentResult,
    config: KonaGlobalConfig,
    cluster_name: str | None,
    items: list[dict[str, Any]],
) -> None:
    resolved = _resolve_cluster_targets(config, cluster_name)
    if resolved:
        for name in resolved:
            await _k8s_apply_items_to_cluster(result, config, name, items)
    else:
        await _k8s_apply_items_to_cluster(result, config, None, items)


def _inject_rollout_annotation(items: list[dict[str, Any]], annotation_path: str | None) -> None:
    if not annotation_path:
        return
    keys = annotation_path.split('.')
    timestamp = datetime.now(tz=UTC).isoformat()
    for item in items:
        target = item
        for key in keys:
            target = target.setdefault(key, {})
        target['konata.dev/deployed-at'] = timestamp


async def k8s_apply_manifests(
    result: DeploymentResult,
    config: KonaGlobalConfig,
    path: Path,
    deployment_config: KonaChallengeConfig.ChallengeDeploymentConfig,
    challenges: list,
) -> None:
    for manifest in deployment_config.kubernetes_manifests:
        manifest_context = build_manifest_context(
            config, result, deployment_config, challenges, use_image_digest=manifest.rollout_restart.image
        )
        items = [
            item
            for manifest_path_str in manifest.paths
            for doc in yaml.safe_load_all(render_template((path / manifest_path_str).read_text(), **manifest_context))
            for item in k8s_expand_manifest(doc)
        ]
        _inject_rollout_annotation(items, manifest.rollout_restart.annotation_path)
        await _k8s_apply_items(result, config, manifest.cluster_name, items)


async def k8s_apply_inline_manifests(
    result: DeploymentResult,
    config: KonaGlobalConfig,
    deployment_config: KonaChallengeConfig.ChallengeDeploymentConfig,
    challenges: list,
) -> None:
    for manifest in deployment_config.kubernetes_inline_manifests:
        manifest_context = build_manifest_context(
            config, result, deployment_config, challenges, use_image_digest=manifest.rollout_restart.image
        )
        items = [
            item
            for doc in manifest.documents
            for item in k8s_expand_manifest(render_template_values(doc, **manifest_context))
        ]
        _inject_rollout_annotation(items, manifest.rollout_restart.annotation_path)
        await _k8s_apply_items(result, config, manifest.cluster_name, items)


def _postprocess_image_names(config: KonaGlobalConfig, challenge_config: KonaChallengeConfig) -> None:
    ctx: dict[str, Any] = {'challenges': challenge_config.challenges, 'config': config}
    for image in challenge_config.deployment.images:
        image.name = render_template(image.name, **ctx)


async def deploy_challenge(
    config: KonaGlobalConfig, path: Path, challenge_config: KonaChallengeConfig
) -> DeploymentResult:
    result = DeploymentResult()
    deployment_config = challenge_config.deployment
    _postprocess_image_names(config, challenge_config)

    await docker_build_images(result, config, path, deployment_config)
    challenges = challenge_config.challenges
    await k8s_apply_manifests(result, config, path, deployment_config, challenges)
    await k8s_apply_inline_manifests(result, config, deployment_config, challenges)
    return result



from collections.abc import Awaitable, Callable
from dataclasses import dataclass

from kona.core.sync import SyncResult
from kona.schema.models import KonaGlobalConfig


@dataclass
class AnalysisContext:
    global_config: KonaGlobalConfig
    sync_result: SyncResult


passes: list[Callable[[AnalysisContext], Awaitable[None]]] = []


def analysis_pass(cb: Callable[[AnalysisContext], Awaitable[None]]) -> Callable[[AnalysisContext], Awaitable[None]]:
    passes.append(cb)
    return cb



import hashlib
from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.parse import quote

from httpx import AsyncClient, Timeout
from loguru import logger

from kona.schema.models import KonaChallengeItem, KonaGlobalConfig, KonaRCTFCredentials
from kona.util.http import raise_for_status

from .abc import ExternalProviderABC


class RCTFProvider(ExternalProviderABC):
    def __init__(self, global_config: KonaGlobalConfig, credentials: KonaRCTFCredentials) -> None:
        self.global_config = global_config
        self.credentials = credentials
        self.bearer_token: str | None = None
        self.challenges_on_remote: list[dict] = []

    @property
    def _client(self) -> AsyncClient:
        headers = dict(self.credentials.extra_headers)
        if self.bearer_token is not None:
            headers['Authorization'] = f'Bearer {self.bearer_token}'
        return AsyncClient(
            base_url=str(self.credentials.base_url),
            headers=headers,
            timeout=Timeout(
                timeout=60,
            ),
        )

    async def setup(self) -> None:
        if self.bearer_token is None:
            async with self._client as client:
                r = await client.post(
                    '/api/v1/auth/login',
                    json={
                        'teamToken': self.credentials.team_token.load(global_config=self.global_config),
                    },
                )
                raise_for_status(r)
                self.bearer_token = r.json()['data']['authToken']
                logger.info('Authenticated in rCTF')

        async with self._client as client:
            r = await client.get('/api/v1/admin/challs')
            raise_for_status(r)
            self.challenges_on_remote = r.json()['data']
        logger.info(f'Retrieved {len(self.challenges_on_remote)} rCTF challenges from remote')

    async def _upload_file(self, file: Path) -> dict[str, str]:
        # If its already deployed, return the existing file info
        async with self._client as client:
            r = await client.post(
                '/api/v2/admin/upload/query',
                json={
                    'uploads': [
                        {
                            'name': file.name,
                            'sha256': hashlib.sha256(file.read_bytes()).hexdigest(),
                        }
                    ]
                },
            )
            raise_for_status(r)
            upload_info = r.json()['data']
            if upload_info and upload_info[0]['url']:
                logger.info(f'File {file.name} is already uploaded to rCTF')
                return upload_info[0]

        # Otherwise, upload it
        async with self._client as client:
            logger.info(f'Uploading {file.name} to rCTF')
            r = await client.post(
                '/api/v2/admin/upload',  # v2
                files={
                    'files': (file.name, BytesIO(file.read_bytes()), 'application/binary'),
                },
            )
            raise_for_status(r)
            return r.json()['data'][0]

    async def sync_challenge(
        self, challenge: KonaChallengeItem, attachment_paths: list[Path], rendered_description: str
    ) -> None:
        uploaded_files: list[dict[str, str]] = [
            await self._upload_file(attachment_path) for attachment_path in attachment_paths
        ]
        challenge_dict: dict[str, Any] = {
            'flag': challenge.flags.rctf,
            'name': challenge.name,
            'files': uploaded_files,
            'author': challenge.author,
            'points': {
                'max': challenge.scoring.initial_value,
                'min': challenge.scoring.minimum_value,
            },
            'category': challenge.category,
            'description': rendered_description,
            'tiebreakEligible': challenge.scoring.rctf.eligible_for_tiebreaks,
        }

        if challenge.instancer_config is not None:
            challenge_dict['instancerConfig'] = {
                'challengeIntegrationId': challenge.instancer_config.challenge_integration_id,
                'config': challenge.instancer_config.config,
                'expose': [
                    {
                        'kind': expose.kind.value,
                        'hostPrefix': expose.host_prefix,
                        'containerName': expose.container_name,
                        'containerPort': expose.container_port,
                        'shouldDisplay': expose.should_display,
                        **(({'title': expose.title}) if expose.title is not None else {}),
                    }
                    for expose in challenge.instancer_config.expose
                ],
                'timeoutMilliseconds': challenge.instancer_config.timeout_milliseconds,
            }
        else:
            challenge_dict['instancerConfig'] = None

        # TODO(es3n1n): cleanup previous attachments if changed

        try:
            existing_challenge = next(
                chal for chal in self.challenges_on_remote if chal['id'] == challenge.challenge_id
            )
            is_up_to_date = all(
                existing_challenge[attr] == challenge_dict[attr] for attr in challenge_dict if attr != 'id'
            )
        except StopIteration:
            is_up_to_date = False

        if is_up_to_date:
            logger.info(f'Challenge {challenge.challenge_id} is already up to date in rCTF')
            return

        async with self._client as client:
            r = await client.put(
                f'/api/v2/admin/challs/{quote(challenge.challenge_id)}',  # v2
                json={
                    'data': challenge_dict,
                },
            )
            raise_for_status(r)
            logger.info(f'Challenge {challenge.challenge_id} has been updated in rCTF')



from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory

from loguru import logger

from kona.core.deployment import DeploymentResult, build_manifest_context, deploy_challenge
from kona.core.k8s_manifest_discovery import discover_deployed_endpoints
from kona.core.kubernetes import load_kubeconfig
from kona.core.provide import resolve_attachments, resolve_source_paths
from kona.external.abc import ExternalProviderABC
from kona.external.ctfd import CTFDProvider
from kona.external.rctf import RCTFProvider
from kona.schema.models import KonaChallengeConfig, KonaGlobalConfig
from kona.schema.parsers import try_load_schema
from kona.util.jinja import render_template, render_template_values


@dataclass
class SynchronizedChallenge:
    description: str = ''
    attachments: list[Path] = field(default_factory=list)


@dataclass
class SynchronizedGroup:
    deployment_result: DeploymentResult
    challenges: list[SynchronizedChallenge] = field(default_factory=list)


@dataclass
class SyncResult:
    groups: list[SynchronizedGroup] = field(default_factory=list)


async def sync_challenge(
    result: SyncResult,
    config: KonaGlobalConfig,
    path: Path,
    challenge: KonaChallengeConfig,
    external_providers: list[ExternalProviderABC],
) -> None:
    logger.info(f'Discovered challenge(s) at {path}: {", ".join(chal.challenge_id for chal in challenge.challenges)}')
    if challenge.discovery.skip:
        logger.warning(f'Skipping {path}')
        return

    # Deploy
    deployment_result = await deploy_challenge(config, path, challenge)
    discover_deployed_endpoints(config, challenge, deployment_result)

    group = SynchronizedGroup(
        deployment_result=deployment_result,
    )

    ctx = build_manifest_context(config, deployment_result, challenge.deployment, challenge.challenges)

    # Sync challenge to the providers
    for chal in challenge.challenges:
        chal.resolve_flags(path)
        ctx['challenge'] = chal

        for ep in chal.endpoints:
            ep.endpoint = render_template(ep.endpoint, **ctx)

        if chal.instancer_config is not None:
            chal.instancer_config.challenge_integration_id = render_template(
                chal.instancer_config.challenge_integration_id, **ctx
            )
            chal.instancer_config.config = render_template_values(chal.instancer_config.config, **ctx)
            for expose in chal.instancer_config.expose:
                expose.host_prefix = render_template(expose.host_prefix, **ctx)

        ctx['endpoints_rendered'] = render_template(config.templates.endpoints_text, **ctx)

        out_chal = SynchronizedChallenge()
        out_chal.description = render_template(config.templates.challenge_description, **ctx)
        out_chal.attachments = resolve_source_paths(path, chal.attachments)

        with TemporaryDirectory() as tmp_dir:
            attachment_paths = resolve_attachments(
                path, Path(tmp_dir), chal.attachments, config.attachment_format, chal.challenge_id
            )
            if attachment_paths:
                logger.info(f'Resolved {len(attachment_paths)} attachment(s) for {chal.challenge_id}')

            for provider in external_providers:
                await provider.sync_challenge(chal, attachment_paths, out_chal.description)
                # challenges were updated, refresh the local cache
                await provider.setup()

        group.challenges.append(out_chal)

    result.groups.append(group)


async def try_discover_challenges(
    result: SyncResult,
    path: Path,
    config: KonaGlobalConfig,
    *,
    depth: int = 0,
    is_root: bool = False,
    external_providers: list[ExternalProviderABC],
    challenge_filter: set[Path] | None = None,
) -> None:
    if depth > config.discovery.challenge_folder_depth:
        return

    # Try load challenge schema
    if not is_root:
        challenge_schema = try_load_schema(path, model=KonaChallengeConfig)
        if challenge_schema is not None:
            if challenge_filter is not None and path.resolve() not in challenge_filter:
                logger.debug(f'Skipping {path} (not in --only filter)')
            else:
                await sync_challenge(result, config, path, challenge_schema, external_providers)

    # Look for challenges in nested folders
    for item in path.iterdir():
        try:
            if not item.is_dir():
                continue
        except OSError as err:
            logger.warning(f'Skipping folder {item} due to {err}')
            continue
        await try_discover_challenges(
            result,
            item,
            config,
            depth=depth + 1,
            external_providers=external_providers,
            challenge_filter=challenge_filter,
        )


async def sync(
    root_path: Path, config: KonaGlobalConfig, *, only_challenges: tuple[str, ...] | None = None
) -> SyncResult:
    result = SyncResult()
    external_providers: list[ExternalProviderABC] = []

    # rCTF
    if config.rctf is not None:
        external_providers.append(RCTFProvider(global_config=config, credentials=config.rctf))

    # CTFd
    if config.ctfd is not None:
        external_providers.append(CTFDProvider(global_config=config, credentials=config.ctfd))

    # Setup external providers
    for provider in external_providers:
        await provider.setup()

    # Set the kubeconfig if there's only one available
    if len(config.clusters) == 1:
        load_kubeconfig(config, next(iter(config.clusters.keys())))

    # Resolve challenge filter
    challenge_filter: set[Path] | None = None
    if only_challenges is not None:
        challenge_filter = {(root_path / p).resolve() for p in only_challenges}
        logger.info(f'Filtering to {len(challenge_filter)} challenge(s): {", ".join(only_challenges)}')

    # Discover challenges
    await try_discover_challenges(
        result,
        root_path,
        config,
        external_providers=external_providers,
        is_root=True,
        challenge_filter=challenge_filter,
    )
    return result
